import { DIFF_DELETE, DIFF_INSERT } from 'diff-match-patch';
import { optionsToConfig } from './config';
import { DomIterator } from './domIterator';
import { areNodesEqual, charForNodeName, diffText, getAncestors, isElement, isText, never } from './util';
/**
 * Stringifies a DOM node recursively. Text nodes are represented by their `data`,
 * while all other nodes are represented by a single Unicode code point
 * from the Private Use Area of the Basic Multilingual Plane.
 */
const serialize = (root, config) => [...new DomIterator(root, config)].reduce((text, node) => text + (isText(node) ? node.data : charForNodeName(node.nodeName)), '');
const getLength = (node) => (isText(node) ? node.length : 1);
export function visualDomDiff(oldRootNode, newRootNode, options = {}) {
    // Define config and simple helpers.
    const config = optionsToConfig(options);
    const { addedClass, modifiedClass, removedClass, skipSelf } = config;
    const notSkipSelf = (node) => !skipSelf(node);
    const getDepth = (node, rootNode) => getAncestors(node, rootNode).filter(notSkipSelf).length;
    const isFormattingNode = (node) => isElement(node) && skipSelf(node);
    const getFormattingAncestors = (node, rootNode) => getAncestors(node, rootNode)
        .filter(isFormattingNode)
        .reverse();
    // Input iterators.
    const diffIterator = diffText(serialize(oldRootNode, config), serialize(newRootNode, config))[Symbol.iterator]();
    const oldIterator = new DomIterator(oldRootNode, config);
    const newIterator = new DomIterator(newRootNode, config);
    // Input variables produced by the input iterators.
    let { done: diffDone, value: diffItem } = diffIterator.next();
    let { done: oldDone, value: oldNode } = oldIterator.next();
    let { done: newDone, value: newNode } = newIterator.next();
    let diffOffset = 0;
    let oldOffset = 0;
    let newOffset = 0;
    // Output variables.
    const document = newRootNode.ownerDocument || newRootNode;
    const rootOutputNode = document.createDocumentFragment();
    let oldOutputNode = rootOutputNode;
    let oldOutputDepth = 0;
    let newOutputNode = rootOutputNode;
    let newOutputDepth = 0;
    let removedNode = null;
    let addedNode = null;
    const removedNodes = new Array();
    const addedNodes = new Array();
    const modifiedNodes = new Array();
    const formattingMap = new Map();
    function prepareOldOutput() {
        const depth = getDepth(oldNode, oldRootNode);
        while (oldOutputDepth > depth) {
            /* istanbul ignore if */
            if (!oldOutputNode.parentNode) {
                return never();
            }
            if (oldOutputNode === removedNode) {
                removedNode = null;
            }
            oldOutputNode = oldOutputNode.parentNode;
            oldOutputDepth--;
        }
        /* istanbul ignore if */
        if (oldOutputDepth !== depth) {
            return never();
        }
    }
    function prepareNewOutput() {
        const depth = getDepth(newNode, newRootNode);
        while (newOutputDepth > depth) {
            /* istanbul ignore if */
            if (!newOutputNode.parentNode) {
                return never();
            }
            if (newOutputNode === addedNode) {
                addedNode = null;
            }
            newOutputNode = newOutputNode.parentNode;
            newOutputDepth--;
        }
        /* istanbul ignore if */
        if (newOutputDepth !== depth) {
            return never();
        }
    }
    function appendChild(node) {
        let nodeModified = false;
        /* istanbul ignore if */
        if (oldOutputNode !== newOutputNode || addedNode || removedNode) {
            return never();
        }
        if (isText(node)) {
            const oldFormatting = getFormattingAncestors(oldNode, oldRootNode);
            const newFormatting = getFormattingAncestors(newNode, newRootNode);
            formattingMap.set(node, newFormatting);
            const length = oldFormatting.length;
            if (length !== newFormatting.length) {
                nodeModified = true;
            }
            else {
                for (let i = 0; i < length; ++i) {
                    if (!areNodesEqual(oldFormatting[i], newFormatting[i])) {
                        nodeModified = true;
                        break;
                    }
                }
            }
        }
        if (nodeModified) {
            modifiedNodes.push(node);
        }
        newOutputNode.appendChild(node);
        oldOutputNode = node;
        newOutputNode = node;
        oldOutputDepth++;
        newOutputDepth++;
    }
    function appendOldChild(node) {
        if (!removedNode) {
            removedNode = node;
            removedNodes.push(node);
        }
        if (isText(node)) {
            const oldFormatting = getFormattingAncestors(oldNode, oldRootNode);
            formattingMap.set(node, oldFormatting);
        }
        oldOutputNode.appendChild(node);
        oldOutputNode = node;
        oldOutputDepth++;
    }
    function appendNewChild(node) {
        if (!addedNode) {
            addedNode = node;
            addedNodes.push(node);
        }
        if (isText(node)) {
            const newFormatting = getFormattingAncestors(newNode, newRootNode);
            formattingMap.set(node, newFormatting);
        }
        newOutputNode.appendChild(node);
        newOutputNode = node;
        newOutputDepth++;
    }
    function nextDiff(step) {
        const length = diffItem[1].length;
        diffOffset += step;
        if (diffOffset === length) {
            ;
            ({ done: diffDone, value: diffItem } = diffIterator.next());
            diffOffset = 0;
        }
        else {
            /* istanbul ignore if */
            if (diffOffset > length) {
                return never();
            }
        }
    }
    function nextOld(step) {
        const length = getLength(oldNode);
        oldOffset += step;
        if (oldOffset === length) {
            ;
            ({ done: oldDone, value: oldNode } = oldIterator.next());
            oldOffset = 0;
        }
        else {
            /* istanbul ignore if */
            if (oldOffset > length) {
                return never();
            }
        }
    }
    function nextNew(step) {
        const length = getLength(newNode);
        newOffset += step;
        if (newOffset === length) {
            ;
            ({ done: newDone, value: newNode } = newIterator.next());
            newOffset = 0;
        }
        else {
            /* istanbul ignore if */
            if (newOffset > length) {
                return never();
            }
        }
    }
    // Copy all content from oldRootNode and newRootNode to rootOutputNode,
    // while deduplicating identical content.
    // Difference markers and formatting are excluded at this stage.
    while (!diffDone) {
        if (diffItem[0] === DIFF_DELETE) {
            /* istanbul ignore if */
            if (oldDone) {
                return never();
            }
            prepareOldOutput();
            const length = Math.min(diffItem[1].length - diffOffset, getLength(oldNode) - oldOffset);
            const text = diffItem[1].substring(diffOffset, diffOffset + length);
            appendOldChild(isText(oldNode)
                ? document.createTextNode(text)
                : oldNode.cloneNode(false));
            nextDiff(length);
            nextOld(length);
        }
        else if (diffItem[0] === DIFF_INSERT) {
            /* istanbul ignore if */
            if (newDone) {
                return never();
            }
            prepareNewOutput();
            const length = Math.min(diffItem[1].length - diffOffset, getLength(newNode) - newOffset);
            const text = diffItem[1].substring(diffOffset, diffOffset + length);
            appendNewChild(isText(newNode)
                ? document.createTextNode(text)
                : newNode.cloneNode(false));
            nextDiff(length);
            nextNew(length);
        }
        else {
            /* istanbul ignore if */
            if (oldDone || newDone) {
                return never();
            }
            prepareOldOutput();
            prepareNewOutput();
            const length = Math.min(diffItem[1].length - diffOffset, getLength(oldNode) - oldOffset, getLength(newNode) - newOffset);
            const text = diffItem[1].substring(diffOffset, diffOffset + length);
            if (isText(oldNode) && isText(newNode)) {
                if (oldOutputNode === newOutputNode) {
                    appendChild(document.createTextNode(text));
                }
                else {
                    appendOldChild(document.createTextNode(text));
                    appendNewChild(document.createTextNode(text));
                }
            }
            else if (oldOutputNode === newOutputNode &&
                areNodesEqual(oldNode, newNode)) {
                appendChild(newNode.cloneNode(false));
            }
            else {
                appendOldChild(isText(oldNode)
                    ? document.createTextNode(text)
                    : oldNode.cloneNode(false));
                appendNewChild(isText(newNode)
                    ? document.createTextNode(text)
                    : newNode.cloneNode(false));
            }
            nextDiff(length);
            nextOld(length);
            nextNew(length);
        }
    }
    // Mark up the content which has been removed.
    for (removedNode of removedNodes) {
        const parentNode = removedNode.parentNode;
        let previousSibling = removedNode.previousSibling;
        // Move the delete before inserts.
        while (previousSibling && addedNodes.includes(previousSibling)) {
            parentNode.insertBefore(removedNode, previousSibling);
            previousSibling = removedNode.previousSibling;
        }
        if (previousSibling &&
            previousSibling.lastChild &&
            removedNodes.includes(previousSibling.lastChild)) {
            previousSibling.appendChild(removedNode);
        }
        else {
            const marker = document.createElement('DEL');
            marker.classList.add(removedClass);
            parentNode.insertBefore(marker, removedNode);
            marker.appendChild(removedNode);
        }
    }
    // Mark up the content which has been added.
    for (addedNode of addedNodes) {
        const parentNode = addedNode.parentNode;
        const previousSibling = addedNode.previousSibling;
        if (previousSibling &&
            previousSibling.lastChild &&
            addedNodes.includes(previousSibling.lastChild)) {
            previousSibling.appendChild(addedNode);
        }
        else {
            const marker = document.createElement('INS');
            marker.classList.add(addedClass);
            parentNode.insertBefore(marker, addedNode);
            marker.appendChild(addedNode);
        }
    }
    // Mark up the content which has been modified.
    if (!config.skipModified) {
        for (const modifiedNode of modifiedNodes) {
            const parentNode = modifiedNode.parentNode;
            const previousSibling = modifiedNode.previousSibling;
            if (previousSibling &&
                previousSibling.lastChild &&
                modifiedNodes.includes(previousSibling.lastChild)) {
                previousSibling.appendChild(modifiedNode);
            }
            else {
                const marker = document.createElement('INS');
                marker.classList.add(modifiedClass);
                parentNode.insertBefore(marker, modifiedNode);
                marker.appendChild(modifiedNode);
            }
        }
    }
    // Add formatting.
    for (const [textNode, formattingNodes] of formattingMap) {
        for (const formattingNode of formattingNodes) {
            const parentNode = textNode.parentNode;
            const previousSibling = textNode.previousSibling;
            if (previousSibling &&
                areNodesEqual(previousSibling, formattingNode)) {
                previousSibling.appendChild(textNode);
            }
            else {
                const clonedFormattingNode = formattingNode.cloneNode(false);
                parentNode.insertBefore(clonedFormattingNode, textNode);
                clonedFormattingNode.appendChild(textNode);
            }
        }
    }
    return rootOutputNode;
}
